from typing import List
from flytekit import TaskMetadata, dynamic, task
from flytekit.extras.tasks.shell import subproc_execute
from flytekit.types.file import FlyteFile
from unionbio.config import main_img_fqn, logger
from unionbio.types import Alignment





@task(container_image=main_img_fqn)
def mark_dups(al: Alignment) -> Alignment:
    """
    Identify and remove duplicates from an alignment file using GATK's MarkDuplicates tool.

    This function takes in an alignment file, removes the duplicates and writes out
    a deduped alignment file.

    Args:
        al (Alignment): An alignment object.

    Returns:
        Alignment: An alignment object with the deduped alignment file and deduplication metrics.
    """
    logger.info(f"Marking duplicates for {al}")
    al.alignment.download()
    al.deduped = True
    mets = al.get_metrics_fname()
    al_out = al.get_alignment_fname()

    cmd = [
        "gatk",
        "MarkDuplicates",
        "-I",
        al.alignment.path,
        "-O",
        al_out,
        "-M",
        mets,
    ]
    logger.debug(f"Running command: {cmd}")
    subproc_execute(cmd)

    al.alignment = FlyteFile(path=al_out)
    al.dedup_metrics = FlyteFile(path=mets)
    logger.info(f"Returning: {al}")
    return al


@dynamic
def mark_dups_samples(als: List[Alignment]) -> List[Alignment]:
    return [mark_dups(al=al) for al in als]



import os
from pathlib import Path
from typing import List
from flytekit import task, dynamic, current_context, Resources
from flytekit.types.file import FlyteFile
from flytekit.extras.tasks.shell import subproc_execute
from unionbio.config import logger, main_img_fqn
from unionbio.types import Alignment, Reference, VCF


@task(container_image=main_img_fqn, requests=Resources(cpu="4", mem="10Gi"))
def haplotype_caller(ref: Reference, al: Alignment) -> VCF:
    """
    Call variants using HaplotypeCaller from GATK.

    Args:
        ref (Reference): A reference object containing the reference FASTA file.
        al (Alignment): An alignment object containing the aligned reads.

    Returns:
        VCF: A VCF object containing the called variants.
    """
    logger.info("Calling variants using HaplotypeCaller")
    logger.info(f"Reference: {ref}")
    logger.info(f"Alignment: {al}")
    ref.aggregate()
    logger.debug(f"Reference dir contents: {os.listdir(ref.ref_dir.path)}")
    al.aggregate()
    con_dir = Path(current_context().working_directory)
    vcf_out = VCF(
        sample=al.sample,
        caller="gatk-hc",
    )
    vcf_fn = con_dir.joinpath(vcf_out.get_vcf_fname())
    vcf_idx_fn = con_dir.joinpath(vcf_out.get_vcf_idx_fname())
    hc_cmd = [
        "gatk",
        "HaplotypeCaller",
        "-R",
        str(ref.get_ref_path()),
        "-I",
        str(al.alignment.path),
        "-O",
        str(vcf_fn),
        "-ERC",
        "GVCF",
    ]
    logger.debug("Running command:")
    logger.debug(" ".join(hc_cmd))
    logger.debug(f"Running in dir {con_dir} with contents: {os.listdir(con_dir)}")
    subproc_execute(hc_cmd, cwd=con_dir)
    vcf_out.vcf = FlyteFile(path=str(vcf_fn))
    vcf_out.vcf_idx = FlyteFile(path=str(vcf_idx_fn))
    return vcf_out

@dynamic
def hc_call_samples(ref: Reference, als: List[Alignment]) -> List[VCF]:
    return [haplotype_caller(ref=ref, al=al) for al in als]



from mashumaro.mixins.json import DataClassJSONMixin
from dataclasses import dataclass
from flytekit.types.file import FlyteFile


@dataclass
class Protein(DataClassJSONMixin):
    """
    Represents a protein sequence and its associated metadata.

    This class defines the structure for representing a protein sequence along with attributes
    that describe the associated metadata.

    Attributes:
        name (str): The name or identifier of the protein sequence.
        protein (FlyteFile): A FlyteFile object representing the path to the protein sequence file.
    """

    name: str
    protein: FlyteFile | None = None
    genes: FlyteFile | None = None

    def get_prot_fname(self):
        return f"{self.name}_proteins.fasta"

    def get_genes_fname(self):
        return f"{self.name}_genes.gff"



import os
import shutil
from itertools import chain
from mashumaro.mixins.json import DataClassJSONMixin
from dataclasses import dataclass, fields
from flytekit import current_context
from flytekit.types.file import FlyteFile
from pathlib import Path
from unionbio.config import logger
from unionbio.tasks.helpers import filter_dir


@dataclass
class Alignment(DataClassJSONMixin):
    """
    Represents a SAM (Sequence Alignment/Map) file and its associated sample and report.

    This class defines the structure for representing a SAM file along with attributes
    that describe the associated sample and report.

    Attributes:
        sample (str): The name or identifier of the sample to which the SAM file belongs.
        aligner (str): The name of the aligner used to generate the SAM file.
        alignment (FlyteFile): A FlyteFile object representing the path to the alignment file.
        alignment_report (FlyteFile): A FlyteFile object representing an associated report
            for performance of the aligner.
        sorted (bool): A boolean value indicating whether the SAM file has been sorted.
        deduped (bool): A boolean value indicating whether the SAM file has been deduplicated.
        bqsr_report (FlyteFile): A FlyteFile object representing a report from the Base Quality
            Score Recalibration (BQSR) process.
    """

    sample: str
    aligner: str
    format: str | None = None
    alignment: FlyteFile | None = None
    alignment_idx: FlyteFile | None = None
    alignment_report: FlyteFile | None = None
    sorted: bool | None = None
    deduped: bool | None = None
    recalibrated: bool | None = None
    dedup_metrics: FlyteFile | None = None
    bqsr_report: FlyteFile | None = None

    def _get_state_str(self):
        state = f"{self.sample}_{self.aligner}"
        if self.sorted:
            state += "_sorted"
        if self.deduped:
            state += "_deduped"
        if self.recalibrated:
            state += "_recal"
        return state

    def get_alignment_fname(self):
        return f"{self._get_state_str()}_aligned.{self.format}"

    def get_alignment_idx_fname(self):
        return f"{self._get_state_str()}_aligned.bam.bai"

    def get_report_fname(self):
        return f"{self._get_state_str()}_aligned_report.txt"

    def get_bqsr_fname(self):
        return f"{self._get_state_str()}_bqsr.table"

    def get_metrics_fname(self):
        return f"{self._get_state_str()}_metrics.txt"
    
    def aggregate(self, target: Path = None) -> Path:
        """
        Explicitly aggregate alignment and supporting files into another given directory. 
        If None is provided, the current working is used.

        Args:
            target (Path): The target directory to move the VCF and index files to.

        Returns:
            Path: The target directory containing the VCF and index files.
        """
        target = target or Path(current_context().working_directory)
        logger.info(f"Aggregating alignment files to {target}")
        os.makedirs(target, exist_ok=True)
        for field in fields(self):
            at = getattr(self, field.name)
            if type(at) == FlyteFile:
                at.download()
                np = target.joinpath(Path(at.path).name)
                if not np.exists():
                    shutil.move(at.path, np)
                setattr(self, field.name, FlyteFile(path=np))
        return target

    @classmethod
    def make_all(cls, dir: Path, include: list[str] = ["*.bam*", "*.sam", "*report*"], exclude: list[str] = []) -> list:
        samples = {}
        for fp in filter_dir(dir, include=include, exclude=exclude):
            sample, aligner = fp.stem.split("_")[0:2]

            if sample not in samples:
                samples[sample] = cls(sample=sample, aligner=aligner)

            if "sorted" in fp.name:
                samples[sample].sorted = True
            else:
                samples[sample].sorted = False

            if "deduped" in fp.name:
                samples[sample].deduped = True
            else:
                samples[sample].deduped = False
            
            if "recal" in fp.name:
                samples[sample].recalibrated = True
            else:
                samples[sample].recalibrated = False

            if fp.name.endswith("bam"):
                samples[sample].format = "bam"
                samples[sample].alignment = FlyteFile(path=str(fp))
            elif fp.name.endswith("bam.bai"):
                samples[sample].alignment_idx = FlyteFile(path=str(fp))
            elif "sam" in fp.name:
                samples[sample].format = "sam"
                samples[sample].alignment = FlyteFile(path=str(fp))
            elif "report" in fp.name:
                samples[sample].alignment_report = FlyteFile(path=str(fp))

        logger.info(f"Created following Alignment objects from {dir}: {samples}")
        return list(samples.values())



import os
import logging
from pathlib import Path

# Setup the logger
logger = logging.getLogger(__name__)
console_handler = logging.StreamHandler()
console_handler.setFormatter(
    logging.Formatter("[%(asctime)s %(levelname)s %(name)s] %(message)s")
)
logger.addHandler(console_handler)
logger.setLevel(logging.DEBUG)

# Default paths
remote_ref = "https://github.com/unionai-oss/unionbio/raw/main/tests/assets/references/GRCh38_chr21.fasta"
remote_reads = [
        "https://github.com/unionai-oss/unionbio/raw/main/tests/assets/sequences/subsampled/SRR812824-sub_1.fastq",
        "https://github.com/unionai-oss/unionbio/raw/main/tests/assets/sequences/subsampled/SRR812824-sub_2.fastq",
    ]
remote_rgtag = "@RG\tID:SRR812824\tSM:Sample1\tLB:Library1\tPL:illumina\tPU:SRR812824.2"
remote_sites_vcf = "https://github.com/unionai-oss/unionbio/raw/main/tests/assets/sites/Mills_and_1000G_gold_standard_chr21.indels.hg38.vcf"
remote_sites_idx = "https://github.com/unionai-oss/unionbio/raw/main/tests/assets/sites/Mills_and_1000G_gold_standard_chr21.indels.hg38.vcf.idx"

# Tool config
fastp_cpu = "3"

current_registry = os.getenv("IMAGE_SPEC_REGISTRY", "docker.io/unionbio")
src_rt = Path(__file__).parent.parent

# Image tags
# While tasks can reference imageSpec directly, using the tag allows registering tasks
# from a containerized environment. These also contain the actual unionbio package.
main_img_fqn = "docker.io/unionbio/main:xqiK1JDwMpuBWnRy4COD4A"
folding_img_fqn = "docker.io/unionbio/folding:P0VHG8pidSc7qQCXQ9ChHg"
parabricks_img_fqn = "docker.io/unionbio/parabricks:poSQ0vCfj9DYDclPS4TXkQ"



from datetime import timedelta
from flytekit import workflow, approve, conditional
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from flytekit import map_task
from unionbio.config import ref_loc, seq_dir_pth
from unionbio.tasks.fastqc import fastqc
from unionbio.tasks.fastp import pyfastp
from unionbio.tasks.utils import prepare_raw_samples, check_fastqc_reports
from unionbio.tasks.bowtie2 import bowtie2_align_samples, bowtie2_index
from unionbio.tasks.multiqc import render_multiqc



@workflow
def alignment_wf(seq_dir: FlyteDirectory = seq_dir_pth) -> FlyteFile:
    """
    Run an alignment workflow on FastQ files contained in the configured seq_dir.

    This function performs QC, preprocessing, index generation, and finally aligments
    on FastQ input files present in a preconfigured S3 prefix. It returns a FlyteFile
    of a MultiQC report containing all relevant statistics of the different steps.

    Args:
        seq_dir (FlyteDirectory, optional): The input directory containing sequencing data.
            Defaults to the value of `seq_dir` (if provided).

    Returns:
        FlyteFile: A FlyteFile object representing the output of the alignment workflow.
    """
    # Generate FastQC reports and check for failures
    fqc_dir = fastqc(seq_dir=seq_dir)
    check = check_fastqc_reports(rep_dir=fqc_dir)

    # If the FastQC summary is PASS or WARN then we can proceed with the workflow.
    # If there is at least one FAIL, then the workflow fails.
    samples = (
        conditional("pass-qc")
        .if_((check == "PASS") | (check == "WARN"))
        .then(prepare_raw_samples(seq_dir=seq_dir))
        .else_()
        .fail("One or more samples failed QC.")
    )

    # Map out filtering across all samples and generate indices
    filtered_samples = map_task(pyfastp)(rs=samples)
    approve_filter = approve(
        render_multiqc(fqc=fqc_dir, filt_reps=filtered_samples, sams=[]),
        "filter-approval",
        timeout=timedelta(hours=2),
    )

    bowtie2_idx = bowtie2_index(ref=ref_loc)

    # Require that samples pass QC before potentially expensive index generation
    samples >> approve_filter
    approve_filter >> bowtie2_idx

    # Compare alignment results using two different aligners in a dynamic task
    sams = bowtie2_align_samples(idx=bowtie2_idx, samples=filtered_samples)

    # Generate final multiqc report with stats from all steps
    return render_multiqc(fqc=fqc_dir, filt_reps=filtered_samples, sams=sams)



import biotite.structure.io as bsio
import torch
from Bio import SeqIO
from flytekit import Resources, task
from flytekit.extras.tasks.shell import subproc_execute
from flytekit.types.file import FlyteFile
from transformers import AutoTokenizer, EsmForProteinFolding
from unionbio.config import logger, folding_img_fqn
from unionbio.types import Reads, Protein

# Ok, we are ready to begin and import all tools we will use in this colab:


@task(container_image=folding_img_fqn)
def test_unionbio_install() -> str:
    return "UnionBio package installed successfully."


@task(container_image=folding_img_fqn)
def prodigal_predict(in_seq: Reads) -> Protein:
    """
    Predicts protein sequences from a DNA sequence using Prodigal.
    """
    seq = in_seq.uread
    seq.download()

    prot = Protein(name=in_seq.sample)
    prot_out = prot.get_prot_fname()
    genes_out = prot.get_genes_fname()
    logger.debug(f"Initialized protein object: {prot}")

    cmd = [
        "prodigal",
        "-i",
        seq.path,
        "-a",
        prot_out,
        "-o",
        genes_out,
        "-p",
        "meta",
        "-f",
        "gff",
    ]
    logger.debug(f"Running Prodigal: {' '.join(cmd)}")
    result = subproc_execute(cmd)
    logger.info(
        f"Command exited with code {result.returncode} and stdout {result.output}"
    )

    prot.protein = prot_out
    prot.genes = genes_out
    logger.info(f"Returning protein object: {prot}")

    return prot


@task(container_image=folding_img_fqn, requests=Resources(gpu="1"))
def esm_fold(prot: FlyteFile) -> FlyteFile:
    esmfold = EsmForProteinFolding.from_pretrained(
        "facebook/esmfold_v1",
        low_cpu_mem_usage=True,  # we set this flag to save some RAM during loading
    )
    # esmfold = esmfold.to(device) # -> move ESMFold to the GPU
    esmfold.esm = esmfold.esm.half()  # -> make sure we use a lightweight precision
    esmfold_tokenizer = AutoTokenizer.from_pretrained("facebook/esmfold_v1")

    # 1. Select a predicted protein-coding gene that we want to fold:
    i = 0  # lets just use the first one.
    protein_record = list(SeqIO.parse(prot.path, "fasta"))[i]
    protein_seq = str(protein_record.seq)[:-1]  # remove stop codon

    # 2. Tokenize the gene to make it digestible for ESMFold:
    esmfold_in = esmfold_tokenizer(
        [protein_seq], return_tensors="pt", add_special_tokens=False
    )

    # 3. Feed the tokenized sequence to ESMfold:
    # (in PyTorch's inference_mode to avoid any costly gradient computations)
    with torch.inference_mode():
        esmfold_out = esmfold(**esmfold_in.to("cuda"))
        esmfold_out_pdb = esmfold.output_to_pdb(esmfold_out)[0]

    protein_structure_pdb = "protein_structure.pdb"
    with open(protein_structure_pdb, "w") as f:
        f.write(esmfold_out_pdb)

    protein_structure = bsio.load_structure(
        protein_structure_pdb, extra_fields=["b_factor"]
    )
    print("Generated protein structure: ", protein_structure)

    return FlyteFile(path=protein_structure_pdb)



from time import sleep
from pathlib import Path
from typing import List
from flytekit import kwtypes, TaskMetadata, task, current_context
from flytekit.extras.tasks.shell import OutputLocation, ShellTask, subproc_execute
from flytekit.types.directory import FlyteDirectory
from unionbio.config import main_img_fqn, logger
from unionbio.types import Reads


@task(container_image=main_img_fqn)
def fastqc(reads: List[Reads]) -> FlyteDirectory:
    """
    Perform quality control using FastQC.

    This function takes a FlyteDirectory object containing raw sequencing data, 
    gathers QC metrics using FastQC, and returns a FlyteDirectory object that
    can be crawled with MultiQC to generate a report.

    Args:
        reads (List[Reads]): A list of Reads objects containing raw sequencing data.

    Returns:
        qc (FlyteDirectory): A directory containing fastqc report output.
    """
    pwd = Path(current_context().working_directory)
    indir = pwd.joinpath("fastqc_in")
    outdir = pwd.joinpath("fastqc_out")
    outdir.mkdir()
    for r in reads:
        r.aggregate(target=indir)
    logger.debug(f"Aggregated reads to {indir} with contents: {list(indir.iterdir())}")
    fqc_cmd = [
        "fastqc",
        f"{indir}/*.fastq*",
        "--outdir",
        str(outdir),
    ]
    fqc_cmd_str = " ".join(fqc_cmd)
    logger.debug(f"Running FastQC with command:")
    logger.debug(fqc_cmd_str)
    res = subproc_execute(fqc_cmd_str, shell=True)
    logger.debug(f"FastQC command returned: {res}")
    logger.debug(f"FastQC reports generated at {outdir} with contents: {list(outdir.iterdir())}")
    
    return FlyteDirectory(path=outdir)



import os
import shutil
import subprocess
from pathlib import Path
from typing import List
from flytekit import kwtypes, task, Resources, current_context, TaskMetadata, dynamic
from flytekit.extras.tasks.shell import OutputLocation, ShellTask, subproc_execute
from flytekit.types.file import FlyteFile
from flytekit.types.directory import FlyteDirectory
from unionbio.config import remote_ref, main_img_fqn, logger
from unionbio.types import Reference, Reads, Alignment




@task(
    container_image=main_img_fqn,
    requests=Resources(cpu="4", mem="10Gi"),
    cache=True,
    cache_version=remote_ref,
)
def bwa_index(ref: Reference) -> Reference:
    """Indexes a reference genome using BWA.

    Args:
        ref (Reference): The reference object containing the reference genome.

    Returns:
        Reference: The updated reference object with associated index and metadata.
    """
    ref.aggregate()

    if f"{ref.ref_name}.fai" not in os.listdir(ref.ref_dir.path):
        sam_idx = ["samtools", "faidx", str(ref.get_ref_path())]
        subproc_execute(sam_idx, cwd=ref.ref_dir.path)
    ref_dict = Path(ref.ref_dir.path).joinpath(ref.get_ref_dict_fn())
    if ref_dict not in os.listdir(ref.ref_dir.path):
        logger.debug(f"Generating sequence dictionary {ref_dict} for {ref.ref_name}")
        res = subproc_execute(["samtools", "dict", ref.ref_name, "-o", ref_dict], cwd=ref.ref_dir.path)
        logger.debug(f"Reference dict exists: {ref_dict.exists()}")
    bwa_idx = ["bwa", "index", str(ref.get_ref_path())]
    subproc_execute(bwa_idx, cwd=ref.ref_dir.path)
    logger.debug(f"Indexing complete for {ref.ref_name}")
    logger.debug(f"Reference dir contents: {os.listdir(ref.ref_dir.path)}")
    ref.index_name = ref.ref_name
    ref.indexed_with = "bwa"
    ref.ref_dir = FlyteDirectory(path=ref.ref_dir.path)

    return ref


@task(
    container_image=main_img_fqn,
    requests=Resources(cpu="4", mem="10Gi"),
)
def bwa_align(ref: Reference, reads: Reads, rgtag: str = "") -> Alignment:
    """Aligns reads to a reference genome using BWA.

    Args:
        ref (Reference): The reference object containing the reference genome and index.
        reads (Reads): The reads object containing the reads to align.

    Returns:
        Alignment: The alignment object containing the aligned reads.
    """
    ref.aggregate()
    reads.aggregate()
    if rgtag:
        rgtag = repr(f'{rgtag}')
    else:
        rgtag = repr('@RG\tID:default\tSM:sample\tPL:illumina\tLB:lib1\tPU:unit1')
    al_out = Alignment(
        sample=reads.sample,
        aligner="bwa",
        format="sam",
        sorted=False,
        deduped=False,
    )
    con_dir = current_context().working_directory
    sam_out = Path(con_dir).joinpath(al_out.get_alignment_fname())
    bwa_align = [
        "bwa",
        "mem",
        "-R",
        rgtag,
        str(ref.get_ref_path()),
        str(reads.read1.path),
        str(reads.read2.path),
        ">",
        str(sam_out),
    ]
    cmd_str = " ".join(bwa_align)
    logger.info(f"Running BWA alignment with: {cmd_str}")
    logger.debug(f"Running in {con_dir} with contents: {os.listdir(con_dir)}")
    subproc_execute(cmd_str, shell=True, cwd=con_dir)
    sp = Path(sam_out)
    logger.debug(f"Alignment exists ({sp.exists()}) at {sp.resolve()}")
    al_out.alignment = FlyteFile(path=str(sam_out))
    return al_out

@dynamic(container_image=main_img_fqn)
def bwa_align_samples(idx: Reference, samples: List[Reads]) -> List[Alignment]:
    """
    Process samples through BWA.

    This function takes a Reference object containing a bwa index and a list of
    Reads objects containing filtered sample data. It performs paired-end alignment
    using bwa. It then returns a list of Alignment objects representing the alignment results.

    Args:
        idx (FlyteDirectory): The FlyteDirectory object representing the bowtie2 index.
        samples (List[Reads]): A list of Reads objects containing sample data
            to be processed.

    Returns:
        [List[Alignment]: A list of alignment results (Alignment objects).
    """
    return [bwa_align(ref=idx, reads=sample) for sample in samples]


from unionbio.types.alignment import Alignment
from unionbio.types.variants import VCF
from unionbio.types.reads import Reads
from unionbio.types.protein import Protein
from unionbio.types.reference import Reference


import gzip
import ftplib
import requests
from pathlib import Path
from typing import List
from flytekit.remote import FlyteRemote
from flytekit.configuration import Config
from unionbio.config import logger

def cache_hash(input: str | List[str]) -> str:
    """
    Generate a hash from a URI to use as a cache key.

    Args:
        url (str | List[str]): The URI or list of URIs to hash.

    Returns:
        str: The first 6 characters of the hash of the URI(s).
    """
    if type(input) == list:
        input = "".join(input)
    return str(hash(input))[:6]

def gunzip_file(gzip_file: Path) -> Path:
    # Ensure the input file exists
    if not gzip_file.exists():
        raise FileNotFoundError(f"{gzip_file} not found")

    # Ensure the file has a .gz extension
    if not gzip_file.suffix == ".gz":
        raise ValueError("Input file is not a gzip file")

    # Define the output file path
    output_file = gzip_file.with_suffix("")

    # Open input and output files
    with gzip.open(gzip_file, "rb") as f_in, open(output_file, "wb") as f_out:
        # Decompress and write to output file
        f_out.write(f_in.read())

    return output_file


def get_remote(local=None, config_file=None):
    """
    Get remote configuration settings and return a remote object.

    This function retrieves remote configuration settings, including the local flag and
    a configuration file, and uses them to create and return a remote object.

    Args:
        local (bool, optional): A flag indicating whether to use local settings. If True,
            the function will use local settings; if False, it will use remote settings.
            Defaults to None, which implies the use of default settings.
        config_file (str, optional): The path to a custom configuration file. If provided,
            this file will be used for configuration settings. Defaults to None.

    Returns:
        Remote: A remote object configured with the specified settings.
    """
    return FlyteRemote(
        config=Config.auto(
            config_file=(
                None
                if local
                else config_file
                if config_file is not None
                else str(Path.home() / ".flyte" / "config-sandbox.yaml")
            )
        ),
        default_project="flytesnacks",
        default_domain="development",
    )


def fetch_file(url: str, local_dir: str) -> Path:
    """
    Downloads a file from the specified URL.

    Args:
        url (str): The URL of the file to download.
        local_dir (Path): The directory where you would like this file saved.

    Returns:
        Path: The local path to the file.

    Raises:
        requests.HTTPError: If an HTTP error occurs while downloading the file.
    """
    url_parts = url.split("/")
    host = url_parts[2]
    prot = url_parts[0]
    fname = url_parts[-1]
    remote_dir = "/".join(url_parts[3:-1])
    local_path = Path(local_dir).joinpath(fname)

    if prot == "ftp:":  # FTP
        ftp = ftplib.FTP(host)
        ftp.login()
        ftp.cwd(remote_dir)
        with open(local_path, "wb") as file:
            ftp.retrbinary(f"RETR {fname}", file.write)
        ftp.quit()
    elif prot == "http:" or prot == "https:":  # HTTP
        try:
            response = requests.get(url)
            with open(local_path, "wb") as file:
                file.write(response.content)
        except requests.HTTPError as e:
            print(f"HTTP error: {e}")
            raise e
    return local_path


def filter_dir(dir: Path, include: list[str] = ["*"], exclude: list[str] = []):
    """
    Filter the contents of a directory based on include and exclude patterns.

    Args:
        dir (Path): The directory whose contents you would like to filter.
        include (list[str], optional): A list of wildcarded patterns to include 
            by passing to rglob. Defaults to ["*"].
        exclude (list[str], optional): A list of patterns to exclude. Defaults to [].

    Returns:
        list[Path]: A list of paths to the filtered contents of the directory.
    """
    logger.debug(f"Filtering {dir} with include={include} and exclude={exclude}")
    included = [f for pattern in include for f in dir.rglob(pattern)]
    logger.debug(f"Included: {included}")
    for f in included:
        if not any([f.match(p) for p in exclude]):
            logger.debug(f"Matched {f}")
            yield f
        else:
            logger.debug(f"Excluded {f}")


from flytekit import workflow, LaunchPlan
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from flytekit import map_task, task, dynamic
from typing import List
from unionbio.config import main_img_fqn, remote_reads, remote_ref, remote_sites_vcf, remote_sites_idx
from unionbio.tasks.fastqc import fastqc
from unionbio.tasks.fastp import pyfastp
from unionbio.tasks.utils import prepare_raw_samples, reformat_alignments, fetch_remote_reads, fetch_remote_reference, fetch_remote_sites
from unionbio.tasks.bowtie2 import bowtie2_align_samples, bowtie2_index
from unionbio.tasks.bwa import bwa_align_samples, bwa_index
from unionbio.tasks.multiqc import render_multiqc
from unionbio.tasks.base_recal import recalibrate_samples
from unionbio.tasks.mark_dups import mark_dups_samples
from unionbio.tasks.sort_sam import sort_samples
from unionbio.tasks.haplotype_caller import hc_call_samples
from unionbio.types import VCF


@workflow
def calling_wf(
    ref_url: str = remote_ref,
    reads_urls: List[str] = remote_reads,
    remote_sites_vcf: str = remote_sites_vcf,
    remote_sites_idx: str = remote_sites_idx,
) -> List[VCF]:
    """
    Run an alignment and variant calling workflow on FastQ files.

    This function performs QC, preprocessing, index generation, alignment, and variant calling
    on FastQ input files present in a pre-configured remote URL. It returns a list of VCF objects
    containing the variant calls for each sample.

    Args:
        ref_url (str): The URL of the reference genome to use for alignment.
        reads_urls (List[str]): A list of URLs pointing to the FastQ files to align.
        remote_sites_vcf (str): The URL of the VCF file containing known variant sites.
        remote_sites_idx (str): The URL of the tabix index for the VCF file.

    Returns:
        List[VCF]: A list of VCF objects containing the variant calls for each sample.
    """
    # Fetch remote inputs
    ref = fetch_remote_reference(url=ref_url)
    reads = fetch_remote_reads(urls=reads_urls)
    sites = fetch_remote_sites(sites=remote_sites_vcf, idx=remote_sites_idx)

    # Generate FastQC reports and check for failures
    fqc_out = fastqc(reads=reads)

    # Map out filtering across all samples and generate indices
    filtered_samples = map_task(pyfastp)(rs=reads)

    # Explicitly define task dependencies
    fqc_out >> filtered_samples

    # Generate a bowtie2 index or load it from cache
    idx = bwa_index(ref=ref)

    # Generate alignments using bowtie2
    sams = bwa_align_samples(idx=idx, samples=filtered_samples)

    # Recalibrate & Reformat
    sorted = sort_samples(als=sams)
    deduped = mark_dups_samples(als=sorted)
    recal_sams = recalibrate_samples(als=deduped, sites=sites, ref=idx)
    bams = reformat_alignments(als=recal_sams, to_format='bam')

    # Call Variants
    vcfs = hc_call_samples(ref=idx, als=bams)

    # Generate final multiqc report with stats from all steps
    render_multiqc(fqc=fqc_out, filt_reps=filtered_samples)

    return vcfs



import os
import shutil
from mashumaro.mixins.json import DataClassJSONMixin
from dataclasses import dataclass
from flytekit.types.directory import FlyteDirectory
from flytekit import current_context
from pathlib import Path
from unionbio.tasks.helpers import gunzip_file, fetch_file
from unionbio.config import logger


@dataclass
class Reference(DataClassJSONMixin):
    """
    Represents a reference FASTA and associated index files.

    This class captures a directory containing a reference FASTA and optionally it's associated
    index files.

    Attributes:
        ref_name (str): Name or identifier of the raw sequencing sample.
        ref_dir (FlyteDirectory): Directory containing the reference and any index files.
        index_name (str): Index string to pass to tools requiring it. Some tools require just the
        ref name and assume index files are in the same dir, others require the index name.
        indexed_with (str): Name of tool used to create the index.
    """

    ref_name: str
    ref_dir: FlyteDirectory
    index_name: str | None = None
    indexed_with: str | None = None

    def get_ref_dict_fn(self):
        return Path(self.ref_name).with_suffix(".dict")

    def get_ref_path(self, unzip=True):
        fp = Path(self.ref_dir.path).joinpath(self.ref_name)
        if ".gz" in self.ref_name and unzip:
            unzipped = gunzip_file(fp)
            self.ref_name = unzipped.name
            return unzipped
        else:
            return fp
        
    def aggregate(self, target: Path = None) -> Path:
        """
        Explicitly aggregate the contents of the reference directory into
        another given directory. If None is provided, the current working
        is used.

        Args:
            target (Path): The target directory to move the reference files to.

        Returns:
            Path: The target directory containing the reference files.
        """
        target = target or Path(current_context().working_directory)
        logger.info(f"Aggregating reference files to {target}")
        self.ref_dir.download()
        os.makedirs(target, exist_ok=True)
        for f in os.listdir(self.ref_dir.path):
            fo = Path(self.ref_dir.path).joinpath(f)
            fd = Path(target).joinpath(f)
            shutil.move(fo, fd)
            logger.debug(f"Moved {fo} to {fd}")
        self.ref_dir.path = target
        return target
    
    @classmethod
    def from_remote(cls, url: str):
        ref = fetch_file(url, "/tmp")
        return cls(ref.name, FlyteDirectory(ref.parent))



import time
import streamlit as st
import streamlit.components.v1 as components
from streamlit import session_state as session
from flytekit.configuration import Config
from flytekit.remote import FlyteRemote
from unionbio.workflows.compare_aligners import alignment_wf
#####################################################
# PYTHONPATH=. streamlit run alignment_dashboard.py #
#####################################################



st.set_page_config(
    page_title="fastq alignment",
    page_icon="https://docs.flyte.org/en/latest/_static/flyte_circle_gradient_1_4x4.png",
)


def gate_node_approval(execution):
    if session.qc_check:
        remote.set_signal("filter-approval", execution.id.name, True)
        while True:
            synced_execution = remote.sync_execution(execution, sync_nodes=True)
            try:
                uri = remote.client.get_download_signed_url(
                    native_url=synced_execution.node_executions["n9"].closure.deck_uri
                )
                break
            except Exception:
                pass
            time.sleep(10)

        st.write("Final MultiQC report:")
        components.iframe(uri.signed_url, scrolling=True, height=1000, width=1000)
    else:
        st.write("Not proceeding with the execution of the workflow.")


def samples():
    execution = remote.execute(alignment_wf, inputs={"seq_dir": session.seq_dir})
    while True:
        synced_execution = remote.sync_execution(execution, sync_nodes=True)
        try:
            uri = remote.client.get_download_signed_url(
                native_url=synced_execution.node_executions["n4"].closure.deck_uri
            )
            break
        except Exception:
            pass
        time.sleep(10)

    components.iframe(uri.signed_url, scrolling=True, height=1000, width=1000)

    user_input_form = st.form(key="user_input_form")
    with user_input_form:
        st.checkbox("Does the MultiQC report look satisfactory to you?", key="qc_check")
        user_input_form.form_submit_button(
            "Submit", on_click=gate_node_approval, args=(execution,)
        )


def data_uploader_form():
    fastq_alignment_form = st.form(key="fastq_alignment_form")
    with fastq_alignment_form:
        st.text_input(
            label="What's the location at which the sequencing data is stored?",
            value="",
            key="seq_dir",
        )
        fastq_alignment_form.form_submit_button(label="Submit", on_click=samples)


if __name__ == "__main__":
    remote = FlyteRemote(
        config=Config.for_sandbox(),
        default_project="flytesnacks",
        default_domain="development",
    )
    data_uploader_form()



import os
from pathlib import Path
from flytekit import task, dynamic, current_context
from flytekit.extras.tasks.shell import subproc_execute
from flytekit.types.file import FlyteFile
from unionbio.config import main_img_fqn, logger
from unionbio.types import Reference, Alignment, VCF



@task(container_image=main_img_fqn)
def base_recalibrator(ref: Reference, sites: VCF, al: Alignment) -> Alignment:
    """
    Recalibrate base quality scores using GATK's BaseRecalibrator.

    Args:
        ref (Reference): A reference object containing the reference FASTA file.
        sites (VCF): A VCF object containing known variant sites.
        al (Alignment): An alignment object containing the aligned reads.

    Returns:
        Alignment: An alignment object with the recalibrated alignment file.
    """
    ref.aggregate()
    sites.aggregate()
    al.aggregate()
    con_dir = Path(current_context().working_directory)
    logger.debug(f"Sites obj:\n{sites.sample}\n{sites.caller}\n{sites.vcf.path}\n{sites.vcf_idx.path}")
    recal_fn = al.get_bqsr_fname()

    gen_table_cmd = [
        "gatk",
        "BaseRecalibrator",
        "-I",
        str(al.alignment.path),
        "-R",
        str(ref.get_ref_path()),
        "--known-sites",
        str(sites.vcf.path),
        "-O",
        str(recal_fn),
    ]
    logger.debug("Running GATK BaseRecalibrator with command:")
    logger.debug(" ".join(gen_table_cmd))
    logger.debug(f"Files present in context: {os.listdir(con_dir)}")
    subproc_execute(command=gen_table_cmd, cwd=con_dir)

    al.recalibrated = True
    al_out_fname = con_dir.joinpath(al.get_alignment_fname())
    apply_recal_cmd = [
        "gatk",
        "ApplyBQSR",
        "-I",
        al.alignment.path,
        "-R",
        ref.get_ref_path(),
        "--bqsr-recal-file",
        recal_fn,
        "-O",
        al_out_fname,
    ]
    subproc_execute(command=apply_recal_cmd, cwd=con_dir)

    al.alignment = FlyteFile(path=str(al_out_fname))

    return al

@dynamic
def recalibrate_samples(als: list[Alignment], sites: VCF, ref: Reference) -> list[Alignment]:
    return [base_recalibrator(ref=ref, sites=sites, al=al) for al in als]


import os
import zipfile
import requests
import tarfile
from pathlib import Path
from typing import List
from flytekit import task, current_context
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from flytekit.extras.tasks.shell import subproc_execute
from unionbio.config import main_img_fqn, logger, parabricks_img_fqn, remote_reads, remote_ref, remote_sites_vcf, remote_sites_idx
from unionbio.tasks.helpers import fetch_file, cache_hash
from unionbio.types import Reads, Reference, VCF, Alignment




@task(container_image=main_img_fqn)
def prepare_raw_samples(seq_dir: FlyteDirectory) -> List[Reads]:
    """
    Prepare and process raw sequencing data to create a list of RawSample objects.

    This function processes raw sequencing data located in the specified input directory
    and prepares it to create a list of RawSample objects.

    Args:
        seq_dir (FlyteDirectory): The input directory containing raw sequencing data.

    Returns:
        List[Reads]: A list of Reads objects representing the processed sequencing data.
    """
    seq_dir.download()
    return Reads.make_all(Path(seq_dir))


@task(container_image=main_img_fqn)
def fetch_remote_reference(url: str) -> Reference:
    workdir = current_context().working_directory
    ref_path = fetch_file(url, workdir)
    return Reference(ref_name=str(ref_path.name), ref_dir=FlyteDirectory(path=workdir))


@task(container_image=main_img_fqn)
def fetch_remote_reads(urls: List[str] = remote_reads) -> List[Reads]:
    """
    Fetches remote reads from a list of URLs and returns a list of Reads objects.

    Args:

    Returns:

    """
    workdir = current_context().working_directory
    for url in urls:
        fetch_file(url, workdir)
    return Reads.make_all(Path(workdir))


@task(container_image=main_img_fqn)
def fetch_remote_sites(sites: str, idx: str) -> VCF:
    """
    Fetches remote known sites from a URL and returns a Sites object.

    Args:

    Returns:

    """
    workdir = current_context().working_directory
    sites_path = fetch_file(sites, workdir)
    idx_path = fetch_file(idx, workdir)
    sample_name = Path(sites_path.stem)
    while sample_name.suffix:
        sample_name = sample_name.with_suffix("")
    return VCF(sample=str(sample_name), caller='NA', vcf=FlyteFile(path=str(sites_path)), vcf_idx=FlyteFile(path=str(idx_path)))


@task
def fetch_files(urls: List[str]) -> List[FlyteFile]:
    outfiles = []
    workdir = current_context().working_directory
    for url in urls:
        lpath = fetch_file(url, workdir)
        outfiles.append(FlyteFile(path=lpath))
    return outfiles


@task
def fetch_tarfile(url: str) -> FlyteDirectory:
    """
    Downloads a tar.gz file from the specified URL, extracts its contents, and returns a FlyteDirectory object.

    Args:
        url (str): The URL of the tar.gz file to download.

    Returns:
        FlyteDirectory: A FlyteDirectory object representing the directory where the contents of the tar.gz file were extracted.

    Raises:
        requests.HTTPError: If an HTTP error occurs while downloading the file.
    """
    try:
        response = requests.get(url)
        tar_name = url.split("/")[-1]
        with open(tar_name, "wb") as file:
            file.write(response.content)
    except requests.HTTPError as e:
        print(f"HTTP error: {e}")
        raise e

    working_dir = current_context().working_directory
    out_dir = Path(os.path.join(working_dir, tar_name))
    with tarfile.open(tar_name, "r:gz") as tarf:
        tarf.extractall(out_dir)

    return FlyteDirectory(path=out_dir)


@task
def check_fastqc_reports(rep_dir: FlyteDirectory) -> str:
    """
    Check FastQC reports for errors.

    This function checks FastQC reports for errors and raises an exception if any are found.

    Args:
        rep_dir (FlyteDirectory): The input directory containing FastQC reports.
    """
    rep_dir.download()
    all_zips = list(Path(rep_dir.path).rglob("*fastqc.zip*"))

    for p in all_zips:
        logger.debug(f"Checking {p}")
        with zipfile.ZipFile(p, "r") as zip_file:
            logger.debug(f"{zip_file.filename}")
            logger.debug(f"Archive contains {zip_file.namelist()}")
            with zip_file.open(
                f"{Path(zip_file.filename).stem}/summary.txt"
            ) as summary:
                contents = summary.read().decode("utf-8")
                logger.debug(f"Contents of summary.txt: {contents}")
                if "FAIL" in contents:
                    return "FAIL"
                elif "WARN" in contents:
                    return "WARN"

    return "PASS"


@task(container_image=parabricks_img_fqn)
def compare_bams(in1: FlyteFile, in2: FlyteFile) -> bool:
    """
    Compares two BAM files and returns True if they are identical, False otherwise.

    Args:
        in1 (FlyteFile): The first input BAM file.
        in2 (FlyteFile): The second input BAM file.

    Returns:
        bool: True if the BAM files are identical, False otherwise.
    """
    in1.download()
    in2.download()

    cmp1 = [
        "bam",
        "diff",
        "--in1",
        in1.path,
        "--in2",
        in2.path,
        "--noCigar",
        "--isize",
        "--flag",
        "--mate",
        "--mapQual",
    ]

    result = subproc_execute(cmp1)

    no_out = result.out == "" and result.err == ""

    return no_out


@task(
    container_image=main_img_fqn,
)
def intersect_vcfs(vcf1: VCF, vcf2: VCF) -> VCF:
    """
    Takes the intersection of 2 VCF files and returns a new VCF file to increase
    calling sensitivity.

    Args:
        vcf1 (VCF): The first input VCF object.
        vcf2 (VCF): The second input VCF object.

    Returns:
        VCF: Intersected and zipped VCF object.
    """
    vcf1.aggregate()
    vcf2.aggregate()
    isec_out = VCF(sample=vcf1.sample, caller=f"{vcf1.caller}_{vcf2.caller}_isec")
    fname_out = isec_out.get_vcf_fname()

    cmd = " ".join(
        [
            "bcftools",
            "isec",
            "-n=2",
            "-O",
            "z",
            "-w",
            "1",
            str(vcf1.vcf.path),
            str(vcf2.vcf.path),
            "-o",
            fname_out,
        ]
    )

    subproc_execute(cmd, shell=True)
    isec_out.vcf = FlyteFile(path=fname_out)

    return isec_out

@task(container_image=main_img_fqn) #TODO: generalize
def reformat_alignments(als: List[Alignment], to_format: str) -> List[Alignment]:
    """
    Reformat alignment files to ensure they are in the correct format.

    Args:
        als (List[Alignment]): A list of Alignment objects containing the aligned reads.

    Returns:
        List[Alignment]: A list of Alignment objects with the reformatted alignment files.
    """
    als_out = []
    for al in als:
        al.aggregate()
        if to_format == "bam":
            al.format = "bam"
            al_out_fname = al.get_alignment_fname()
            convert_cmd = [
                "samtools",
                "view",
                "-S",
                "-b",
                str(al.alignment.path),
                "-o",
                al_out_fname,
            ]
            subproc_execute(convert_cmd)
            logger.debug("Running reformat cmd:")
            logger.debug(' '.join(convert_cmd))
            al.alignment = FlyteFile(path=al_out_fname)
            logger.debug(f"Reformatted alignment file {al_out_fname} exists: {os.path.exists(al_out_fname)}")
            
            idx_fn = al.get_alignment_idx_fname()
            idx_cmd = ["samtools", "index", al_out_fname, '-o', idx_fn]
            logger.debug("Running index cmd:")
            logger.debug(' '.join(idx_cmd))
            subproc_execute(idx_cmd)
            logger.debug(f"Index file {idx_fn} exists: {os.path.exists(idx_fn)}")
            al.alignment_idx = FlyteFile(path=idx_fn)
        als_out.append(al)
        
    return als_out


import os
from pathlib import Path
from typing import List
from flytekit import kwtypes, task, Resources, current_context, TaskMetadata, dynamic
from flytekit.extras.tasks.shell import OutputLocation, ShellTask, subproc_execute
from flytekit.types.file import FlyteFile
from flytekit.types.directory import FlyteDirectory
from unionbio.config import main_img_fqn, logger
from unionbio.types import Alignment, Reads, Reference




@task(
    container_image=main_img_fqn,
    requests=Resources(cpu="4", mem="10Gi"),
)
def bowtie2_index(ref: Reference) -> Reference:
    ref.aggregate()
    if f"{ref.ref_name}.fai" not in os.listdir(ref.ref_dir.path):
        logger.debug(f"Samtools indexing {ref.ref_name}")
        subproc_execute(["samtools", "faidx", ref.ref_name], cwd=ref.ref_dir.path)
    ref_dict = ref.get_ref_dict_fn()
    if ref_dict not in os.listdir(ref.ref_dir.path):
        logger.debug(f"Generating sequence dictionary {ref_dict} for {ref.ref_name}")
        res = subproc_execute(["samtools", "dict", ref.ref_name, "-o", ref_dict], cwd=ref.ref_dir.path)
        logger.debug(f"Reference dict exists: {ref_dict.exists()}")
    ref.index_name = "bt2_idx"
    ref.indexed_with = "bowtie2"
    idx_cmd = [
        "bowtie2-build",
        ref.ref_name,
        ref.index_name
    ]
    logger.debug(f"Running command: {idx_cmd}")
    subproc_execute(idx_cmd, cwd=ref.ref_dir.path)
    logger.debug(f"Index created at {ref.ref_dir.path} with contents {os.listdir(ref.ref_dir.path)}")
    ref.ref_dir = FlyteDirectory(path=ref.ref_dir.path)
    return ref


@task(
    container_image=main_img_fqn,
    requests=Resources(cpu="4", mem="10Gi"),
)
def bowtie2_align_paired_reads(idx: Reference, fs: Reads) -> Alignment:
    """
    Perform paired-end alignment using Bowtie 2 on a filtered sample.

    This function takes a Reference object containing the Bowtie 2 index and a
    FiltSample object containing filtered sample data. It performs paired-end alignment
    using Bowtie 2 and returns a Alignment object representing the resulting alignment.

    Args:
        idx (Reference): A Reference object containing the Bowtie 2 index.
        fs (Reads): A filtered sample Reads object containing filtered sample data to be aligned.

    Returns:
        Alignment: An Alignment object representing the alignment result.
    """
    idx.aggregate()
    fs.aggregate()
    logger.debug(f"Index downloaded to {idx.ref_dir.path}")
    alignment = Alignment(fs.sample, "bowtie2", "sam", sorted=False, deduped=False)
    al = Path(alignment.get_alignment_fname()).resolve()
    rep = Path(alignment.get_report_fname()).resolve()
    logger.debug(f"Writing alignment to {al} and report to {rep}")

    cmd = [
        "bowtie2",
        "-x",
        idx.index_name,
        "-1",
        str(fs.read1.path),
        "-2",
        str(fs.read2.path),
        "-S",
        str(al),
    ]

    logger.debug(f"Running command: {cmd}")
    result = subproc_execute(cmd, cwd=idx.ref_dir.path)
    logger.debug(f"Alignment produced: {al.exists()}")
    with open(rep, "w") as f:
        f.write(result.error)

    alignment.alignment = FlyteFile(path=str(al))
    alignment.alignment_report = FlyteFile(path=str(rep))
    return alignment


@dynamic(container_image=main_img_fqn)
def bowtie2_align_samples(idx: Reference, samples: List[Reads]) -> List[Alignment]:
    """
    Process samples through bowtie2.

    This function takes a FlyteDirectory objects representing a bowtie index and a list of
    Reads objects containing filtered sample data. It performs paired-end alignment
    using bowtie2. It then returns a list of Alignment objects representing the alignment results.

    Args:
        bt2_idx (FlyteDirectory): The FlyteDirectory object representing the bowtie2 index.
        samples (List[Reads]): A list of Reads objects containing sample data
            to be processed.

    Returns:
        List[List[Alignment]]: A list of lists, where each inner list contains alignment
            results (Alignment objects) for a sample, with results from both aligners.
    """
    sams = []
    for sample in samples:
        sam = bowtie2_align_paired_reads(idx=idx, fs=sample)
        sams.append(sam)
    return sams


from typing import List
from flytekit import TaskMetadata, dynamic, task
from flytekit.extras.tasks.shell import subproc_execute
from flytekit.types.file import FlyteFile
from unionbio.config import main_img_fqn, logger
from unionbio.types import Alignment




@task(container_image=main_img_fqn)
def sort_sam(al: Alignment) -> Alignment:
    """
    Sort an alignment file using GATK's SortSam tool.

    Args:
        al (Alignment): An alignment object.

    Returns:
        Alignment: An alignment object with the sorted alignment file.
    """
    logger.info(f"Sorting: {al}")
    al.alignment.download()
    al.sorted = True
    al_out = al.get_alignment_fname()

    cmd = [
        "gatk",
        "SortSam",
        "-I",
        al.alignment.path,
        "-O",
        al_out,
        "-SO",
        "coordinate",
    ]
    logger.debug(f"Running command: {cmd}")
    subproc_execute(cmd)

    al.alignment = FlyteFile(path=al_out)
    logger.info(f"Returning: {al}")
    return al


@dynamic
def sort_samples(als: List[Alignment]) -> List[Alignment]:
    return [sort_sam(al=al) for al in als]



from pathlib import Path
from flytekit import task, Resources, current_context
from flytekit.types.file import FlyteFile
from flytekit.extras.tasks.shell import subproc_execute
from unionbio.config import main_img_fqn, logger, fastp_cpu
from unionbio.types import Reads



@task(
    requests=Resources(cpu=fastp_cpu, mem="2Gi"),
    container_image=main_img_fqn,
)
def pyfastp(rs: Reads) -> Reads:
    """
    Perform quality filtering and preprocessing using Fastp on a RawSample.

    This function takes a RawSample object containing raw sequencing data, performs quality
    filtering and preprocessing using the pyfastp tool, and returns a FiltSample object
    representing the filtered and processed data.

    Args:
        rs (RawSample): A RawSample object containing raw sequencing data to be processed.

    Returns:
        FiltSample: A FiltSample object representing the filtered and preprocessed data.
    """
    ldir = Path(current_context().working_directory)
    samp = Reads(rs.sample)
    samp.filtered = True
    o1, o2 = samp.get_read_fnames()
    rep = samp.get_report_fname()
    o1p = ldir.joinpath(o1)
    o2p = ldir.joinpath(o2)
    repp = ldir.joinpath(rep)
    logger.debug(f"Writing filtered reads to {o1p} and {o2p} and report to {repp}")

    cmd = [
        "fastp",
        "-i",
        rs.read1,
        "-I",
        rs.read2,
        "--thread",
        str(int(fastp_cpu) * 2),
        "-o",
        o1p,
        "-O",
        o2p,
        "-j",
        repp,
    ]
    logger.debug(f"Running command: {cmd}")

    subproc_execute(cmd)

    samp.read1 = FlyteFile(path=str(o1p))
    samp.read2 = FlyteFile(path=str(o2p))
    samp.filt_report = FlyteFile(path=str(repp))

    return samp



from typing import List
from flytekit import workflow
from unionbio.tasks.utils import (
from unionbio.tasks.bwa import bwa_index
from unionbio.tasks.parabricks import pb_fq2bam, pb_deepvar, pb_haplocall
from unionbio.types import VCF

    fetch_remote_reads,
    fetch_remote_reference,
    fetch_remote_sites,
    intersect_vcfs,
)


@workflow
def wgs_small_var_calling_wf(
    reads: List[str] = [
        "https://storage.googleapis.com/brain-genomics-public/research/sequencing/fastq/hiseqx/wgs_pcr_free/30x/HG002.hiseqx.pcr-free.30x.R1.fastq.gz",
        "https://storage.googleapis.com/brain-genomics-public/research/sequencing/fastq/hiseqx/wgs_pcr_free/30x/HG002.hiseqx.pcr-free.30x.R2.fastq.gz",
    ],
    ref: str = "ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCA/000/001/405/GCA_000001405.15_GRCh38/seqs_for_alignment_pipelines.ucsc_ids/GCA_000001405.15_GRCh38_no_alt_analysis_set.fna.gz",
    sites: List[str] = [
        "https://storage.googleapis.com/genomics-public-data/resources/broad/hg38/v0/Mills_and_1000G_gold_standard.indels.hg38.vcf.gz",
        "https://storage.googleapis.com/genomics-public-data/resources/broad/hg38/v0/Mills_and_1000G_gold_standard.indels.hg38.vcf.gz.tbi",
    ],
) -> VCF:
    read_obj = fetch_remote_reads(urls=reads)
    ref_obj = fetch_remote_reference(url=ref)
    sites_obj = fetch_remote_sites(sites=sites[0], idx=sites[1])
    ref_idx = bwa_index(ref_obj=ref_obj)
    alignment = pb_fq2bam(reads=read_obj, sites=sites_obj, ref=ref_idx)
    deepvar_vcf = pb_deepvar(al=alignment, ref=ref_idx)
    haplocall_vcf = pb_haplocall(al=alignment, ref=ref_idx)
    return intersect_vcfs(vcf1=deepvar_vcf, vcf2=haplocall_vcf)



from typing import Tuple
from pathlib import Path
from flytekit import task, Resources
from flytekit.extras.tasks.shell import subproc_execute
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from unionbio.config import parabricks_img_fqn
from unionbio.types import Alignment, Reference, Reads, VCF




@task(requests=Resources(gpu="1", mem="32Gi", cpu="32"), container_image=parabricks_img_fqn)
def pb_fq2bam(reads: Reads, sites: VCF, ref: Reference) -> Alignment:
    """
    Takes an input directory containing paired-end FASTQ files and an indexed reference genome and
    performs alignment using Parabricks' accelerated version of BWA-MEM.

    Args:
        reads (Reads): The input directory containing paired-end FASTQ files.
        sites (VCF): The input VCF file containing known sites.
        ref (Reference): The indexed reference genome.

    Returns:
        Alignment: The resulting alignment in Alignment format.
    """
    RGTAG = "@RG\tID:HG002\tLB:lib\tPL:Illumina\tSM:HG002\tPU:HG002"
    reads.read1.download()
    reads.read2.download()
    sites.vcf.download()
    sites.vcf_idx.download()
    ref.ref_dir.download()

    al_out = Alignment(sample=reads.sample, aligner="pbrun_fq2bam")

    bam_out = al_out.get_alignment_fname()
    bam_idx_out = al_out.get_alignment_idx_fname()
    recal_out = al_out.get_bqsr_fname()

    out, err = subproc_execute(
        [
            "pbrun",
            "fq2bam",
            "--ref",
            str(ref.get_ref_path()),
            "--in-fq",
            str(reads.read1.path),
            str(reads.read2.path),
            RGTAG,
            "--knownSites",
            str(sites[0].path),
            "--out-bam",
            bam_out,
            "--out-recal-file",
            recal_out,
        ]
    )

    al_out.alignment = FlyteFile(path=bam_out)
    al_out.alignment_index = FlyteFile(path=bam_idx_out)
    al_out.recal_file = FlyteFile(path=recal_out)

    return FlyteFile(path=bam_out), FlyteFile(path=recal_out)


@task(requests=Resources(gpu="1", mem="32Gi", cpu="32"), container_image=parabricks_img_fqn)
def basic_align(indir: FlyteDirectory) -> Tuple[FlyteFile, str]:
    """
    Aligns paired-end sequencing reads using BWA-MEM and GATK tools, and returns the path to the processed BAM file
    and the elapsed time for the alignment process.

    Args:
        indir (FlyteDirectory): The input directory containing sequencing read files and reference data.

    Returns:
        FlyteFile: The path to the processed BAM file in FlyteFile format.
    """
    indir.download()
    loc_dir = Path(indir.path)
    r1 = loc_dir.joinpath("Data/sample_1.fq.gz")
    r2 = loc_dir.joinpath("Data/sample_2.fq.gz")
    ref = loc_dir.joinpath("Ref/Homo_sapiens_assembly38.fasta")
    sites = loc_dir.joinpath("Ref/Homo_sapiens_assembly38.known_indels.vcf.gz")
    bampath = "bwa_mem_out.bam"

    cmd = " ".join(
        [
            "bwa",
            "mem",
            "-t",
            "32",
            "-K",
            "10000000",
            "-R",
            r'"@RG\tID:sample_rg1\tLB:lib1\tPL:bar\tSM:sample\tPU:sample_rg1"',
            str(ref),
            str(r1),
            str(r2),
            "|",
            "java",
            "-jar",
            "/usr/local/bin/gatk",
            "SortSam",
            "--MAX_RECORDS_IN_RAM",
            "5000000",
            "-I",
            "/dev/stdin",
            "-O",
            bampath,
            "--SORT_ORDER",
            "coordinate",
        ]
    )

    out1, err1 = subproc_execute(cmd, shell=True)

    dup_bam = "mark_dups.bam"
    out2, err2 = subproc_execute(
        [
            "java",
            "-jar",
            "/usr/local/bin/gatk",
            "MarkDuplicates",
            "-I",
            bampath,
            "-O",
            dup_bam,
            "-M",
            "metrics.txt",
        ]
    )

    recal_out = "recal_data.table"
    out3, err3 = subproc_execute(
        [
            "java",
            "-jar",
            "/usr/local/bin/gatk",
            "BaseRecalibrator",
            "--input",
            dup_bam,
            "--output",
            recal_out,
            "--known-sites",
            str(sites),
            "--reference",
            str(ref),
        ]
    )

    return FlyteFile(path=dup_bam)


@task(requests=Resources(gpu="1", mem="32Gi", cpu="32"), container_image=parabricks_img_fqn)
def pb_deepvar(al: Alignment, ref: Reference) -> VCF:
    """
    Takes an input directory containing BAM files and an indexed reference genome and
    performs variant calling using Parabricks' deepvariant tool.

    Args:

    Returns:

    """
    ref.ref_dir.download()
    al.alignment.download()
    al.alignment_idx.download()

    vcf_out = VCF(sample=al.sample, caller="pbrun_deepvariant")
    vcf_fname = vcf_out.get_vcf_fname()
    vcf_idx_fname = vcf_out.get_vcf_idx_fname()

    deepvar_result = subproc_execute(
        [
            "pbrun",
            "deepvariant",
            "--ref",
            str(ref.get_ref_path()),
            "--in-bam",
            str(al.alignment.path),
            "--out-variants",
            vcf_out,
        ]
    )

    vcf_out.vcf = FlyteFile(path=vcf_fname)
    vcf_out.vcf_idx = FlyteFile(path=vcf_idx_fname)

    deepvar_dir = FlyteDirectory(path="/tmp")
    return deepvar_dir


@task(requests=Resources(gpu="1", mem="32Gi", cpu="32"), container_image=parabricks_img_fqn)
def pb_haplocall(al: Alignment, ref: Reference) -> VCF:
    """
    Takes an input directory containing BAM files and an indexed reference genome and
    performs variant calling using Parabricks' accelerated version of GATK's HaplotypeCaller.

    Args:

    Returns:

    """
    ref.ref_dir.download()
    al.alignment.download()
    al.alignment_idx.download()
    al.bqsr_report.download()

    vcf_out = VCF(sample=al.sample, caller="pbrun_deepvariant")
    vcf_fname = vcf_out.get_vcf_fname()
    vcf_idx_fname = vcf_out.get_vcf_idx_fname()

    out, err = subproc_execute(
        [
            "pbrun",
            "deepvariant",
            "--ref",
            str(ref.get_ref_path()),
            "--in-bam",
            str(al.alignment.path),
            "--in-recal-file",
            str(al.bqsr_report.path),
            "--out-variants",
            vcf_out,
        ]
    )

    vcf_out.vcf = FlyteFile(path=vcf_fname)
    vcf_out.vcf_idx = FlyteFile(path=vcf_idx_fname)

    deepvar_dir = FlyteDirectory(path="/tmp")
    return deepvar_dir



import gzip
import shutil
from pathlib import Path
from flytekit import kwtypes, task, Resources, current_context, TaskMetadata
from flytekit.extras.tasks.shell import OutputLocation, ShellTask, subproc_execute
from flytekit.types.file import FlyteFile
from flytekit.types.directory import FlyteDirectory
from unionbio.config import remote_ref, main_img_fqn, logger
from unionbio.types import Alignment, Reads



"""
Generate Hisat2 index files from a reference genome.

Args:
    ref (FlyteFile): A FlyteFile object representing the input reference file.

Returns:
    FlyteDirectory: A FlyteDirectory object containing the index files.
"""
hisat2_index = ShellTask(
    name="hisat2-index",
    debug=True,
    metadata=TaskMetadata(retries=3, cache=True, cache_version=remote_ref),
    requests=Resources(cpu="4", mem="10Gi"),
    container_image=main_img_fqn,
    script="""
    mkdir {outputs.idx}
    hisat2-build {inputs.ref} {outputs.idx}/hs2_idx
    """,
    inputs=kwtypes(ref=FlyteFile),
    output_locs=[
        OutputLocation(var="idx", var_type=FlyteDirectory, location="/tmp/hs2_idx")
    ],
)


@task(
    container_image=main_img_fqn,
    requests=Resources(cpu="4", mem="10Gi"),
)
def hisat2_align_paired_reads(idx: FlyteDirectory, fs: Reads) -> Alignment:
    """
    Perform paired-end alignment using Hisat 2 on a filtered sample.

    This function takes a FlyteDirectory object representing the Hisat 2 index and a
    Reads object containing filtered sample data. It performs paired-end alignment
    using Hisat 2 and returns a Alignment object representing the resulting alignment.

    Args:
        idx (FlyteDirectory): A FlyteDirectory object representing the Hisat 2 index.
        fs (Reads): A Reads object containing filtered sample data to be aligned.

    Returns:
        Alignment: An Alignment object representing the alignment result.
    """
    idx.download()
    ldir = Path(current_context().working_directory)
    alignment = Alignment(fs.sample, "hisat2", "sam")
    al = ldir.joinpath(alignment.get_alignment_fname())
    rep = ldir.joinpath(alignment.get_report_fname())
    logger.debug(f"Writing alignment to {al} and report to {rep}")

    unc_r1 = ldir.joinpath(f"{fs.sample}_1.fq")
    unc_r2 = ldir.joinpath(f"{fs.sample}_2.fq")

    with gzip.open(fs.read1, "rb") as gz_file, open(unc_r1, "wb") as out_file:
        shutil.copyfileobj(gz_file, out_file)
    with gzip.open(fs.read2, "rb") as gz_file, open(unc_r2, "wb") as out_file:
        shutil.copyfileobj(gz_file, out_file)
    logger.debug(f"Uncompressed reads to {unc_r1} and {unc_r2}")

    cmd = [
        "hisat2",
        "-x",
        f"{idx.path}/hs2_idx",
        "-1",
        unc_r1,
        "-2",
        unc_r2,
        "-S",
        al,
        "--summary-file",
        rep,
    ]
    logger.debug(f"Running command: {cmd}")

    result = subproc_execute(cmd)
    logger.info(
        f"Hisat exited with code {result.returncode}, output: {result.output}, error: {result.error}"
    )

    alignment.alignment = FlyteFile(path=str(al))
    alignment.alignment_report = FlyteFile(path=str(rep))
    alignment.sorted = False
    alignment.deduped = False

    return alignment



import os
import shutil
from mashumaro.mixins.json import DataClassJSONMixin
from dataclasses import dataclass
from flytekit.types.file import FlyteFile
from flytekit import current_context
from pathlib import Path
from unionbio.config import logger
from unionbio.tasks.helpers import filter_dir


@dataclass
class Reads(DataClassJSONMixin):
    """
    Represents a sequencing reads sample via its associated fastq files.

    This class defines the structure for representing a sequencing sample. It includes
    attributes for the sample name and paths to the read files (R1 and R2).

    Attributes:
        sample (str): The name or identifier of the raw sequencing sample.
        filtered (bool): A boolean value indicating whether the reads have been filtered.
        filt_report (FlyteFile): A FlyteFile object representing the path to the filter report.
        read1 (FlyteFile): A FlyteFile object representing the path to the raw R1 read file.
        read2 (FlyteFile): A FlyteFile object representing the path to the raw R2 read file.
    """

    sample: str
    filtered: bool | None = None
    filt_report: FlyteFile | None = None
    uread: FlyteFile | None = None
    read1: FlyteFile | None = None
    read2: FlyteFile | None = None

    def get_read_fnames(self):
        filt = "filt." if self.filtered else ""
        return (
            f"{self.sample}_1.{filt}fastq.gz",
            f"{self.sample}_2.{filt}fastq.gz",
        )

    def get_report_fname(self):
        return f"{self.sample}_fastq-filter-report.json"
    
    def aggregate(self, target: Path = None) -> Path:
        """
        Explicitly aggregate Reads files into another given directory. 
        If None is provided, the current working is used.

        Args:
            target (Path): The target directory to move the reference files to.

        Returns:
            Path: The target directory containing the reference files.
        """
        target = target or Path(current_context().working_directory)
        logger.info(f"Aggregating Reads files to {target}")
        os.makedirs(target, exist_ok=True)
        if self.uread:
            self.uread.download()
            np = target.joinpath(Path(self.uread.path).name)
            if not np.exists():
                shutil.move(self.uread.path, target)
                self.uread.path = np
        elif self.read1 and self.read2:
            self.read1.download()
            self.read2.download()
            np1 = target.joinpath(Path(self.read1.path).name)
            np2 = target.joinpath(Path(self.read2.path).name)
            if not all([np1.exists(), np2.exists()]):
                shutil.move(self.read1.path, np1)
                shutil.move(self.read2.path, np2)
            self.read1.path = np1
            self.read2.path = np2
        else:
            logger.error("No read files found to aggregate!")
        return target

    @classmethod
    def make_all(cls, dir: Path, include: list[str] = ["*fast*"], exclude: list[str] = []):
        samples = {}
        for fp in filter_dir(dir, include=include, exclude=exclude):
            logger.debug(f"Processing {fp}")
            fn = fp.name
            pfn = Path(fn)
            sufs = fp.suffixes
            while pfn.suffix:
                pfn = pfn.with_suffix("")
            fns = str(pfn).split("_")
            sample = fns[0]
            logger.debug(f"Found sample {sample}")

            if sample not in samples:
                samples[sample] = cls(sample=sample)

            if any(s in sufs for s in [".fastq", ".fasta", ".fq"]):
                mate = fns[-1]
                logger.debug(f"Found mate {mate} for {sample}")
                if "1" in mate:
                    samples[sample].read1 = FlyteFile(path=str(fp))
                elif "2" in mate:
                    samples[sample].read2 = FlyteFile(path=str(fp))
                else:
                    samples[sample].uread = FlyteFile(path=str(fp))
            elif "filter-report" in fn and ".json" in sufs:
                logger.debug(f"Found filter report for {sample}")
                samples[sample].filtered = True
                samples[sample].filt_report = FlyteFile(path=str(fp))

        logger.info(f"Created {samples} from {dir}")
        return list(samples.values())



import os
import shutil
from flytekit import ImageSpec, current_context, task
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from flytekit.extras.tasks.shell import subproc_execute
from typing import List
from pathlib import Path
from unionbio.config import logger, main_img_fqn
from unionbio.types import Alignment, Reads, VCF




@task(container_image=main_img_fqn, enable_deck=True)
def render_multiqc(
    fqc: FlyteDirectory, 
    filt_reps: List[Reads] | None, 
    sams: List[Alignment] | None,
    vcfs: List[VCF] | None,
) -> FlyteFile:
    """
    Generate MultiQC report by rendering quality and alignment data.

    Takes a FlyteDirectory object containing FastQC reports (`fqc`),
    a list of FiltSample objects containing reports (`filt_reps`), and a
    list of lists of SamFile objects representing alignment results (`sams`). It generates
    a MultiQC report by rendering quality and alignment data and returns a FlyteFile object
    of the report.

    Args:
        fqc (FlyteDirectory): A FlyteDirectory object containing FastQC reports.
        filt_reps (List[FiltSample]): A list of FiltSample objects representing filtered samples.
        sams (List[List[SamFile]]): A list of lists of SamFile objects representing alignment results.

    Returns:
        FlyteFile: A FlyteFile object representing the MultiQC report.
    """
    ldir = Path(current_context().working_directory)

    fqc.download()
    for f in os.listdir(fqc.path):
        src = ldir.joinpath(fqc.path, f)
        dest = ldir.joinpath(ldir, f)
        shutil.move(src, dest)
    logger.debug(f"FastQC reports downloaded to {ldir}")

    if filt_reps:
        for filt_rep in filt_reps:
            filt_rep.filt_report.download()
            shutil.move(filt_rep.filt_report.path, ldir)
        logger.debug(f"FastP reports downloaded to {ldir}")

    if sams:
        for sam in sams:
            sam.alignment_report.download()
            shutil.move(sam.alignment_report.path, ldir)
        logger.debug(f"Alignment reports for {sams} downloaded to {ldir}")

    final_report = ldir.joinpath("multiqc_report.html")
    mqc_cmd = ["multiqc", str(ldir), "-n", str(final_report)]
    logger.debug(f"Generating MultiQC report at {final_report} with command: {mqc_cmd}")
    subproc_execute(mqc_cmd)

    # Hack to force render plots on page load
    report_html_lst = []
    hack = """
  // Force render plots on page load
  $(".hc-plot").each(function () {
    var target = $(this).attr("id");
    var max_num = mqc_config["num_datasets_plot_limit"] * 50;
    plot_graph(target, undefined, max_num);
  });
  """
    with open(final_report, "r") as f:
        for line in f:
            if "// Render plots on page load" in line:
                report_html_lst.append(hack)
                report_html_lst.append(line)
            else:
                report_html_lst.append(line)
    report_html = "".join(report_html_lst)

    current_context().default_deck.append(report_html)
    logger.debug("MultiQC report HTML added to default deck")

    return FlyteFile(path=str(final_report))



import os
import shutil
from pathlib import Path
from mashumaro.mixins.json import DataClassJSONMixin
from dataclasses import dataclass
from flytekit import current_context
from flytekit.types.file import FlyteFile
from unionbio.config import logger
from unionbio.tasks.helpers import filter_dir


@dataclass
class VCF(DataClassJSONMixin):
    """
    Represents a VCF (Variant Call Format) file and its associated sample and index.

    This class defines the structure for representing a VCF file along with attributes
    that describe the associated sample and index.

    Attributes:
        sample (str): The name or identifier of the sample to which the VCF file belongs.
        caller (str): The name of the variant caller used to generate the VCF.
        vcf (FlyteFile): The VCF file.
        vcf_idx (FlyteFile): The index file for the VCF.

    """

    sample: str
    caller: str
    vcf: FlyteFile | None = None
    vcf_idx: FlyteFile | None = None

    def _get_state_str(self):
        state = f"{self.sample}_{self.caller}"
        return state

    def get_vcf_fname(self):
        return f"{self._get_state_str()}.g.vcf.gz"

    def get_vcf_idx_fname(self):
        return f"{self._get_state_str()}.g.vcf.gz.tbi"

    def aggregate(self, target: Path = None) -> Path:
        """
        Explicitly aggregate VCF and index into another given directory. 
        If None is provided, the current working is used.

        Args:
            target (Path): The target directory to move the VCF and index files to.

        Returns:
            Path: The target directory containing the VCF and index files.
        """
        target = target or Path(current_context().working_directory)
        logger.info(f"Aggregating VCF and index to {target}")
        os.makedirs(target, exist_ok=True)
        self.vcf.download()
        np1 = target.joinpath(Path(self.vcf.path).name)
        if not np1.exists():
            shutil.move(self.vcf.path, np1)
        self.vcf.path = np1
        
        if self.vcf_idx:
            self.vcf_idx.download()
            np2 = target.joinpath(Path(self.vcf_idx.path).name)
            if not np2.exists():
                shutil.move(self.vcf_idx.path, np2)
            self.vcf_idx.path = np2

    @classmethod
    def make_all(cls, dir: Path, include: list[str] = ["*.vcf*"], exclude: list[str] = []) -> list:
        samples = {}
        for fp in filter_dir(dir, include=include, exclude=exclude):
            stem = str(fp.stem).split(".")[0]
            sample, caller = stem.split("_")[0:2]

            if sample not in samples:
                samples[sample] = cls(sample=sample, caller=caller)

            if "tbi" in fp.name or "idx" in fp.name:
                samples[sample].vcf_idx = FlyteFile(path=str(fp))
            if "vcf" in fp.name and "tbi" not in fp.name and "idx" not in fp.name:
                samples[sample].vcf = FlyteFile(path=str(fp))

        logger.info(f"Created following VCF objects from {dir}: {samples}")
        return list(samples.values())
